# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = (
    "CalculationRowIterable",
    "ChainRowIterable",
    "MaterializedRowIterable",
    "ProjectionRowIterable",
    "RowIterable",
    "RowMapping",
    "RowSequence",
    "SelectionRowIterable",
    "ChainRowIterable",
)

import itertools
from abc import abstractmethod
from collections.abc import Callable, Iterator, Mapping, Sequence, Set
from typing import Any

from .._columns import ColumnTag


class RowIterable:
    """An abstract base class for iterables that use mappings for rows.

    `RowIterable` is the `~.Relation.payload` type for the `.iteration` engine.
    """

    @abstractmethod
    def __iter__(self) -> Iterator[Mapping[ColumnTag, Any]]:
        raise NotImplementedError()

    def to_mapping(self, unique_key: Sequence[ColumnTag]) -> RowMapping:
        """Convert this iterable to a `RowMapping`, unless it already is one.

        Parameters
        ----------
        unique_key : `~collections.abc.Sequence` [ `ColumnTag` ]
            Sequence of columns to extract into a `tuple` to use as keys in the
            mapping, guaranteeing uniqueness over these columns.

        Returns
        -------
        rows : `RowMapping`
            A `RowIterable` backed by a mapping.
        """
        return RowMapping(unique_key, {tuple(row[k] for k in unique_key): row for row in self})

    def to_sequence(self) -> RowSequence:
        """Convert this iterable to a `RowSequence`, unless it already is one.

        Returns
        -------
        rows : `RowSequence`
            A `RowIterable` backed by a sequence.
        """
        return RowSequence(list(self))

    def materialized(self) -> MaterializedRowIterable:
        """Convert this iterable to one that holds its rows in a Python
        collection of some kind, instead of generating them lazily.

        Returns
        -------
        rows : `MaterializedRowIterable`
            A `RowIterable` that isn't lazy.
        """
        return self.to_sequence()

    def sliced(self, start: int, stop: int | None) -> RowIterable:
        """Apply a slice operation to this `RowIterable`.

        Parameters
        ----------
        start : `int`
            Start index.
        stop : `int` or `None`
            Stop index (one-past-the-end), or `None` to include up through the
            last row.

        Returns
        -------
        rows : `RowIterable`
            Iterable representing the slice.  May or may not be lazy.
        """
        return SliceRowIterable(self, start, stop)


class MaterializedRowIterable(RowIterable):
    """A `RowIterable` that is not lazy and has a known length."""

    @abstractmethod
    def __len__(self) -> int:
        raise NotImplementedError()

    def materialized(self) -> MaterializedRowIterable:
        # Docstring inherited.
        return self


class RowMapping(MaterializedRowIterable):
    """A `RowIterable` backed by a `~collections.abc.Mapping`

    Parameters
    ----------
    unique_key : `~collections.abc.Sequence` [ `ColumnTag` ]
        Sequence of columns to extract into a `tuple` to use as keys in the
        mapping, guaranteeing uniqueness over these columns.
    rows : `collections.abc.Mapping`
        Nested mapping with `tuple` keys and row values, where each row is
        (as usual for `RowIterable` types) itself a `Mapping` with `.ColumnTag`
        keys.
    """

    def __init__(self, unique_key: Sequence[ColumnTag], rows: Mapping[tuple, Mapping[ColumnTag, Any]]):
        self.rows = rows
        self.unique_key = unique_key

    def __iter__(self) -> Iterator[Mapping[ColumnTag, Any]]:
        return iter(self.rows.values())

    def __len__(self) -> int:
        return len(self.rows)

    def to_mapping(self, unique_key: Sequence[ColumnTag]) -> RowMapping:
        # Docstring inherited.
        if unique_key == self.unique_key:
            return self
        else:
            return super().to_mapping(unique_key)


class RowSequence(MaterializedRowIterable):
    """A `RowIterable` backed by a `~collections.abc.Sequence`

    Parameters
    ----------
    rows : `Mapping`
        Sequence of rows, where each row is (as usual for `RowIterable` types)
        a `Mapping` with `.ColumnTag` keys.
    """

    def __init__(self, rows: Sequence[Mapping[ColumnTag, Any]]):
        self.rows = rows

    def __iter__(self) -> Iterator[Mapping[ColumnTag, Any]]:
        return iter(self.rows)

    def __len__(self) -> int:
        return len(self.rows)

    def to_sequence(self) -> RowSequence:
        # Docstring inherited.
        return self

    def sliced(self, start: int, stop: int | None) -> RowIterable:
        # Docstring inherited.
        return RowSequence(self.rows[start:stop])


class CalculationRowIterable(RowIterable):
    """A `RowIterable` implementation that implements a calculation operation.

    Parameters
    ----------
    target : `RowIterable`
        Original iterable.
    tag : `ColumnTag`
        Key for the new column in result-row mappings.
    callable : `Callable`
        Callable that takes a single mapping argument and returns a new column
        value.
    """

    def __init__(
        self, target: RowIterable, tag: ColumnTag, callable: Callable[[Mapping[ColumnTag, Any]], Any]
    ):
        self.target = target
        self.tag = tag
        self.callable = callable

    def __iter__(self) -> Iterator[Mapping[ColumnTag, Any]]:
        return ({**row, self.tag: self.callable(row)} for row in self.target)


class ChainRowIterable(RowIterable):
    """A `RowIterable` implementation that wraps `itertools.chain`.

    Parameters
    ----------
    chain : `Sequence` [ `RowIterable` ]
        Sequence of iterables to chain together.
    """

    def __init__(self, chain: Sequence[RowIterable]):
        self.chain = chain

    def __iter__(self) -> Iterator[Mapping[ColumnTag, Any]]:
        return itertools.chain.from_iterable(self.chain)


class ProjectionRowIterable(RowIterable):
    """A `RowIterable` implementation that implements a projection operation.

    Parameters
    ----------
    target : `RowIterable`
        Original iterable to take a column subset from.
    columns : `Set`
        Columns to include in the new iterable.
    """

    def __init__(self, target: RowIterable, columns: Set[ColumnTag]):
        self.target = target
        self.columns = columns

    def __iter__(self) -> Iterator[Mapping[ColumnTag, Any]]:
        return ({k: row[k] for k in self.columns} for row in self.target)


class SelectionRowIterable(RowIterable):
    """A `RowIterable` implementation that implements a selection operation.

    Parameters
    ----------
    target : `RowIterable`
        Original iterable to filter rows from.
    callable : `Callable`
        Callable that takes a single mapping argument and returns a `bool`.
    """

    def __init__(self, target: RowIterable, callable: Callable[[Mapping[ColumnTag, Any]], bool]):
        self.target = target
        self.callable = callable

    def __iter__(self) -> Iterator[Mapping[ColumnTag, Any]]:
        return (row for row in self.target if self.callable(row))


class SliceRowIterable(RowIterable):
    """A `RowIterable` that implements a lazy `Slice` operation.

    Parameters
    ----------
    target : `RowIterable`
        Original iterable.
    start : `int`
        Start index.
    stop : `int` or `None`
        Stop index (one-past-the-end), or `None` to include up through the
        last row.
    """

    def __init__(self, target: RowIterable, start: int, stop: int | None):
        self.target = target
        self.start = start
        self.stop = stop

    def __iter__(self) -> Iterator[Mapping[ColumnTag, Any]]:
        for n, row in enumerate(self.target):
            if self.stop is not None and n == self.stop:
                return
            if n >= self.start:
                yield row



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Projection",)

import dataclasses
from collections.abc import Set
from typing import TYPE_CHECKING, Literal, final

from .._columns import ColumnTag
from .._exceptions import ColumnError
from .._operation_relations import UnaryOperationRelation
from .._unary_operation import Identity, UnaryCommutator, UnaryOperation

if TYPE_CHECKING:
    from .._engine import Engine
    from .._relation import Relation


@final
@dataclasses.dataclass(frozen=True)
class Projection(UnaryOperation):
    """A unary operation that removes one or more columns.

    Notes
    -----
    This is the only operation permitted to introduce duplication among rows
    (as opposed to just propagating duplicates).
    """

    columns: frozenset[ColumnTag]
    """The columns to be kept (`frozenset` [ `ColumnTag` ]).
    """

    @property
    def columns_required(self) -> Set[ColumnTag]:
        # Docstring inherited.
        return self.columns

    @property
    def is_empty_invariant(self) -> Literal[True]:
        # Docstring inherited.
        return True

    @property
    def is_count_invariant(self) -> Literal[True]:
        # Docstring inherited.
        return True

    def __str__(self) -> str:
        return f"Π[{', '.join(sorted(str(tag) for tag in self.columns))}]"

    def _begin_apply(
        self, target: Relation, preferred_engine: Engine | None
    ) -> tuple[UnaryOperation, Engine]:
        if self.columns == target.columns:
            return Identity(), target.engine
        if not self.columns <= target.columns:
            raise ColumnError(
                f"Cannot project column(s) {set(self.columns) - target.columns} "
                f"that are not present in the target relation {target}."
            )
        return super()._begin_apply(target, preferred_engine)

    def _finish_apply(self, target: Relation) -> Relation:
        if self.columns == target.columns:
            return target
        return super()._finish_apply(target)

    def applied_columns(self, target: Relation) -> Set[ColumnTag]:
        # Docstring inherited.
        return self.columns

    def applied_min_rows(self, target: Relation) -> int:
        # Docstring inherited.
        return target.min_rows

    def commute(self, current: UnaryOperationRelation) -> UnaryCommutator:
        # Docstring inherited.
        from ._calculation import Calculation

        commuted_columns: frozenset[ColumnTag] = self.columns
        match current.operation:
            case Projection():
                # We can just drop any existing Projection as this one
                # supersedes it; by construction the new one has a
                # subset of the original's columns.
                return UnaryCommutator(first=self, second=Identity())
            case Calculation(tag=tag):
                if tag not in self.columns:
                    # Projection will drop the column added by the
                    # Calculation, so it might as well have never
                    # existed.
                    return UnaryCommutator(first=self, second=Identity())
                else:
                    commuted_columns -= {tag}
        if not commuted_columns >= current.operation.columns_required:
            # Can't move the entire projection past this operation;
            # move what we can, and return the full Projection as the
            # "remainder".
            return UnaryCommutator(
                first=Projection(commuted_columns | current.operation.columns_required),
                second=current.operation,
                done=False,
                messages=(
                    f"{current.operation} requires columns "
                    f"{set(current.operation.columns_required - self.columns)}",
                ),
            )
        return UnaryCommutator(Projection(commuted_columns), current.operation)

    def simplify(self, upstream: UnaryOperation) -> UnaryOperation | None:
        # Docstring inherited.
        from ._calculation import Calculation

        # See similar checks in commute for explanations.
        match upstream:
            case Projection():
                return self
            case Calculation(tag=tag) if tag not in self.columns:
                return self
        return None



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("LeafRelation",)

import dataclasses
from collections.abc import Sequence, Set
from typing import TYPE_CHECKING, Any, Literal, final

from ._relation import BaseRelation

if TYPE_CHECKING:
    from ._columns import ColumnTag
    from ._engine import Engine


@final
@dataclasses.dataclass(frozen=True)
class LeafRelation(BaseRelation):
    """A `Relation` class that represents direct storage of rows, rather than
    an operation on some other relation.
    """

    engine: Engine = dataclasses.field(repr=False, compare=True)
    """The engine that is responsible for interpreting this relation
    (`Engine`).
    """

    columns: frozenset[ColumnTag] = dataclasses.field(repr=False, compare=True)
    """The columns in this relation (`~collections.abc.Set` [ `ColumnTag` ] ).
    """

    payload: Any = dataclasses.field(repr=False, compare=False)
    """The engine-specific contents of the relation."""

    name: str = dataclasses.field(repr=True, compare=True, default="")
    """Name used to identify and reconstruct this relation (`str`)."""

    name_prefix: dataclasses.InitVar[str | None] = "leaf"
    """Prefix used when calling `Engine.get_relation_name` when `name` is not
    provided (`str`).
    """

    messages: Sequence[str] = dataclasses.field(repr=False, compare=False, default=())
    """Messages for use when processing the relation with the `Diagnostics`
    class or similar algorithms (`~collections.abc.Sequence` [ `str` ]).

    This is typically used to explain why a leaf relation has no rows when
    ``max_rows==0``; see `make_doomed`.
    """

    parameters: Any = dataclasses.field(repr=True, compare=True, default=None)
    """Extra data used to uniquely identify and/or reconstruct this relation.
    """

    min_rows: int = dataclasses.field(repr=False, compare=False, default=0)
    """The minimum number of rows this relation might have (`int`)."""

    max_rows: int | None = dataclasses.field(repr=False, compare=False, default=None)
    """The maximum number of rows this relation might have (`int` or `None`).
    """

    @property
    def is_locked(self) -> Literal[True]:
        """Whether this relation and those upstream of it should be considered
        fixed by tree-manipulation algorithms (`bool`).

        See `Relation.is_locked`.
        """
        return True

    def __post_init__(self, name_prefix: str = "leaf") -> None:
        if not self.name:
            object.__setattr__(self, "name", self.engine.get_relation_name(name_prefix))
        if self.max_rows is not None and self.max_rows < self.min_rows:
            raise ValueError(f"max_rows ({self.max_rows}) < min_rows ({self.min_rows})")

    @classmethod
    def make_doomed(
        cls, engine: Engine, columns: Set[ColumnTag], messages: Sequence[str], name: str = "0"
    ) -> LeafRelation:
        """Construct a leaf relation with no rows and one or more messages
        explaining why.

        Parameters
        ----------
        engine : `Engine`
            The engine that is responsible for interpreting this relation.
        columns : `~collections.abc.Set` [ `ColumnTag` ]
            The columns in this relation.
        messages : `~collections.abc.Sequence [ `str` ]
            One or more messages explaining why the relation has no rows.
        name : `str`, optional
            Name used to identify and reconstruct this relation.

        Returns
        -------
        relation : `LeafRelation`
            Doomed leaf relation.
        """
        return LeafRelation(
            engine=engine,
            columns=frozenset(columns),
            min_rows=0,
            max_rows=0,
            payload=engine.get_doomed_payload(columns),
            name=name,
            messages=messages,
        )

    @classmethod
    def make_join_identity(cls, engine: Engine, name: str = "I") -> LeafRelation:
        """Construct a leaf relation with no columns and exactly one row.

        Parameters
        ----------
        engine : `Engine`
            The engine that is responsible for interpreting this relation.
        name : `str`, optional
            Name used to identify and reconstruct this relation.

        Returns
        -------
        relation : `LeafRelation`
            Leaf relation with no columns and one row.
        """
        return LeafRelation(
            engine=engine,
            columns=frozenset(),
            min_rows=1,
            max_rows=1,
            payload=engine.get_join_identity_payload(),
            name=name,
        )

    def __str__(self) -> str:
        return self.name if self.parameters is None else f"{self.name}({self.parameters})"



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Calculation",)

import dataclasses
from collections.abc import Set
from typing import TYPE_CHECKING, Literal, final

from .._columns import ColumnExpression, ColumnTag
from .._exceptions import ColumnError
from .._operation_relations import UnaryOperationRelation
from .._unary_operation import UnaryCommutator, UnaryOperation

if TYPE_CHECKING:
    from .._engine import Engine
    from .._relation import Relation


@final
@dataclasses.dataclass(frozen=True)
class Calculation(UnaryOperation):
    """A relation operation that adds a new column from an expression involving
    existing columns.

    Notes
    -----
    `Calculation` operations are assumed to be deterministically related to
    existing columns - in particular, a `Deduplication` is assumed to have the
    same effect regardless of whether it is performed before or after a
    `Calculation`.  This means a `Calculation` should not be used to generate
    random numbers or counters, though it does not prohibit additional
    information outside the relation being used.  The expression that backs
    a `Calculation` must depend on at least one existing column, however; it
    also cannot be used to add a constant-valued column to a relation.
    """

    tag: ColumnTag
    """Identifier for the new column (`ColumnTag`).
    """

    expression: ColumnExpression
    """Expression used to populate the new column (`ColumnExpression`).
    """

    def __post_init__(self) -> None:
        if not self.expression.columns_required:
            # It's unlikely anyone would want them, and explicitly prohibiting
            # calculated columns that are constants saves us from having to
            # worry about one-row, zero-column relations hiding behind them,
            # and hence Relation.is_trivial not propagating the way we'd like.
            raise ColumnError(
                f"Calculated column {self.tag} that does not depend on any other columns is not allowed."
            )

    @property
    def columns_required(self) -> Set[ColumnTag]:
        # Docstring inherited.
        return self.expression.columns_required

    @property
    def is_empty_invariant(self) -> Literal[True]:
        # Docstring inherited.
        return True

    @property
    def is_count_invariant(self) -> Literal[True]:
        # Docstring inherited.
        return True

    def __str__(self) -> str:
        return f"+[{self.tag!s}={self.expression!s}]"

    def is_supported_by(self, engine: Engine) -> bool:
        # Docstring inherited.
        return self.expression.is_supported_by(engine)

    def _begin_apply(
        self, target: Relation, preferred_engine: Engine | None
    ) -> tuple[UnaryOperation, Engine]:
        # Docstring inherited.
        if not (self.expression.columns_required <= target.columns):
            raise ColumnError(
                f"Cannot calculate column {self.tag} because expression requires "
                f"columns {set(self.expression.columns_required) - target.columns} "
                f"that are not present in the target relation {target}."
            )
        if self.tag in target.columns:
            raise ColumnError(f"Calculated column {self.tag} is already present in {target}.")
        return super()._begin_apply(target, preferred_engine)

    def applied_columns(self, target: Relation) -> Set[ColumnTag]:
        # Docstring inherited.
        result = set(target.columns)
        result.add(self.tag)
        return result

    def applied_min_rows(self, target: Relation) -> int:
        # Docstring inherited.
        return target.min_rows

    def commute(self, current: UnaryOperationRelation) -> UnaryCommutator:
        # Docstring inherited.
        from ._projection import Projection

        if not self.columns_required <= current.target.columns:
            return UnaryCommutator(
                first=None,
                second=current.operation,
                done=False,
                messages=(
                    f"{current.target} is missing columns "
                    f"{set(self.columns_required - current.target.columns)}",
                ),
            )
        # If we commute a calculation before a projection, the
        # projection also needs to include the calculated column.
        return UnaryCommutator(
            self,
            (
                Projection(current.operation.columns | {self.tag})
                if isinstance(current.operation, Projection)
                else current.operation
            ),
        )



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Payload",)

import dataclasses
from typing import Generic, TypeVar

import sqlalchemy

from .._columns import ColumnTag

_L = TypeVar("_L")


@dataclasses.dataclass(eq=False)
class Payload(Generic[_L]):
    """A struct that represents a SQL table or simple ``SELECT`` query via
    SQLAlchemy objects.
    """

    from_clause: sqlalchemy.sql.FromClause
    """SQLAlchemy representation of the FROM clause or table
    (`sqlalchemy.sql.FromClause`).
    """

    where: list[sqlalchemy.sql.ColumnElement] = dataclasses.field(default_factory=list)
    """SQLAlchemy representation of the WHERE clause, as a sequence of
    boolean expressions to be combined with ``AND``
    (`Sequence` [ `sqlalchemy.sql.ColumnElement` ]).
    """

    columns_available: dict[ColumnTag, _L] = dataclasses.field(default_factory=dict)
    """Mapping from `.ColumnTag` to logical column for the columns available
    from the FROM clause (`dict`).
    """

    def copy(self) -> Payload[_L]:
        """Return a copy of this struct that can safely be modified.

        This method takes care to copy mutable attributes while leaving
        immutable objects alone.
        """
        return dataclasses.replace(
            self, where=list(self.where), columns_available=dict(self.columns_available)
        )



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = (
    "SortTerm",
    "Sort",
)

import dataclasses
from collections.abc import Set
from typing import TYPE_CHECKING, final

from .._columns import ColumnTag
from .._exceptions import ColumnError
from .._operation_relations import UnaryOperationRelation
from .._unary_operation import Identity, Reordering, UnaryCommutator, UnaryOperation

if TYPE_CHECKING:
    from .._columns import ColumnExpression
    from .._engine import Engine
    from .._relation import Relation


@dataclasses.dataclass
class SortTerm:
    expression: ColumnExpression
    ascending: bool = True

    def __str__(self) -> str:
        return f"{'' if self.ascending else '-'}{self.expression}"


@final
@dataclasses.dataclass(frozen=True)
class Sort(Reordering):
    """A relation operation that orders rows according to a sequence of
    column expressions.
    """

    terms: tuple[SortTerm, ...] = ()
    """Criteria for sorting rows (`Sequence` [ `SortTerm` ])."""

    @property
    def columns_required(self) -> Set[ColumnTag]:
        # Docstring inherited.
        result: set[ColumnTag] = set()
        for term in self.terms:
            result.update(term.expression.columns_required)
        return result

    def __str__(self) -> str:
        return f"sort[{', '.join(str(term) for term in self.terms)}]"

    def is_supported_by(self, engine: Engine) -> bool:
        # Docstring inherited.
        return all(term.expression.is_supported_by(engine) for term in self.terms)

    def _begin_apply(
        self, target: Relation, preferred_engine: Engine | None
    ) -> tuple[UnaryOperation, Engine]:
        # Docstring inherited.
        if not self.terms:
            return Identity(), target.engine
        for term in self.terms:
            if not term.expression.columns_required <= target.columns:
                raise ColumnError(
                    f"Sort term {term} for target relation {target} needs "
                    f"columns {set(term.expression.columns_required - target.columns)}."
                )
        return super()._begin_apply(target, preferred_engine)

    def _finish_apply(self, target: Relation) -> Relation:
        # Docstring inherited.
        if not self.terms:
            return target
        return super()._finish_apply(target)

    def then(self, next: Sort) -> Sort:
        """Compose this sort with another one.

        Parameters
        ----------
        next : `Sort`
            Sort that acts after ``self``.

        Returns
        -------
        composition : `Sort`
            Sort that is equivalent to ``self`` and ``next`` being applied
            back-to-back.
        """
        new_terms = list(next.terms)
        for term in self.terms:
            if term not in new_terms:
                new_terms.append(term)
        return Sort(tuple(new_terms))

    def commute(self, current: UnaryOperationRelation) -> UnaryCommutator:
        # Docstring inherited.
        if not self.columns_required <= current.target.columns:
            return UnaryCommutator(
                first=None,
                second=current.operation,
                done=False,
                messages=(
                    f"{current.target} is missing columns "
                    f"{set(self.columns_required - current.target.columns)}",
                ),
            )
        if current.operation.is_order_dependent:
            return UnaryCommutator(
                first=None,
                second=current.operation,
                done=False,
                messages=(f"{current.operation} is order-dependent",),
            )
        return UnaryCommutator(self, current.operation)

    def simplify(self, upstream: UnaryOperation) -> UnaryOperation | None:
        # Docstring inherited.
        if not self.terms:
            return upstream
        match upstream:
            case Sort():
                return upstream.then(self)
        return None



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the relations of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = ("Chain",)

import dataclasses
from collections.abc import Set
from typing import TYPE_CHECKING, final

from .._binary_operation import BinaryOperation
from .._columns import ColumnTag
from .._exceptions import ColumnError, EngineError

if TYPE_CHECKING:
    from .._relation import Relation


@final
@dataclasses.dataclass(frozen=True)
class Chain(BinaryOperation):
    """A relation operation that concatenates the rows of a pair of relations
    with the same columns."""

    def __str__(self) -> str:
        return "∪"

    def _begin_apply(self, lhs: Relation, rhs: Relation) -> BinaryOperation:
        # Docstring inherited.
        if lhs.engine != rhs.engine:
            raise EngineError(f"Mismatched chain engines: {lhs.engine} != {rhs.engine}.")
        if lhs.columns != rhs.columns:
            raise ColumnError(f"Mismatched chain columns: {set(lhs.columns)} != {set(rhs.columns)}.")
        return self

    def applied_columns(self, lhs: Relation, rhs: Relation) -> Set[ColumnTag]:
        # Docstring inherited.
        return lhs.columns

    def applied_min_rows(self, lhs: Relation, rhs: Relation) -> int:
        # Docstring inherited.
        return lhs.min_rows + rhs.min_rows

    def applied_max_rows(self, lhs: Relation, rhs: Relation) -> int | None:
        # Docstring inherited.
        return None if lhs.max_rows is None or rhs.max_rows is None else lhs.max_rows + rhs.max_rows



# This file is part of daf_relation.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import unittest

from lsst.daf.relation import ColumnExpression, Materialization, iteration, tests


class MaterializationTestCase(tests.RelationTestCase):
    """Tests for the Materialization operation and relations based on it."""

    def setUp(self) -> None:
        self.a = tests.ColumnTag("a")
        self.engine = iteration.Engine(name="preferred")
        self.leaf = self.engine.make_leaf(
            {self.a}, payload=iteration.RowSequence([{self.a: 0}, {self.a: 1}]), name="leaf"
        )
        # Materializing a leaf just returns the leaf, so we add a Selection to
        # make applying it nontrivial.
        self.target = self.leaf.with_rows_satisfying(
            ColumnExpression.reference(self.a).gt(ColumnExpression.literal(0))
        )

    def test_attributes(self) -> None:
        """Check that all UnaryOperation and Relation attributes have the
        expected values.
        """
        relation = self.target.materialized(name_prefix="prefix")
        assert isinstance(relation, Materialization)
        self.assertEqual(relation.columns, {self.a})
        self.assertEqual(relation.engine, self.engine)
        self.assertEqual(relation.min_rows, self.target.min_rows)
        self.assertEqual(relation.max_rows, self.target.max_rows)
        self.assertTrue(relation.is_locked)
        self.assertTrue(relation.name.startswith("prefix"))

    def test_apply_simplify(self) -> None:
        """Test that applying a Materialization to a leaf or an existing
        materialization does nothing.
        """
        self.assertEqual(self.leaf.materialized(), self.leaf)
        self.assertEqual(self.target.materialized(name="a").materialized("b"), self.target.materialized("a"))

    def test_iteration(self) -> None:
        """Test Materialization execution in the iteration engine."""
        relation = self.target.materialized(name="m")
        self.assertEqual(
            list(self.engine.execute(relation)),
            [{self.a: 1}],
        )
        self.assertIsNotNone(relation.payload)
        self.assertIsInstance(relation.payload, iteration.MaterializedRowIterable)

    def test_str(self) -> None:
        """Test str(Materialization) and
        str(UnaryOperationRelation[Materialization]).
        """
        relation = self.target.materialized(name="m")
        self.assertEqual(str(relation), f"materialize['m']({self.target})")


if __name__ == "__main__":
    unittest.main()
